import json
import re

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.bearlib import deprecate_settings


@linter(executable='remark',
        use_stdin=True,
        use_stdout=True,
        use_stderr=True)
class MarkdownBear:
    """
    Check and correct Markdown style violations automatically.

    See <https://github.com/wooorm/remark-lint> for details about the tool
    below.
    """

    LANGUAGES = {'Markdown'}
    REQUIREMENTS = {NpmRequirement('remark-cli', '2'),
                    NpmRequirement('remark-lint', '5')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}

    _output_regex = re.compile(
        r'\s*(?P<line>\d+):(?P<column>\d+)'
        r'\s*(?P<severity>warning)\s*(?P<message>.*)')

    @staticmethod
    @deprecate_settings(bullets='markdown_bullets',
                        closed_headings='markdown_closed_headings',
                        setext_headings='markdown_setext_headings',
                        emphasis='markdown_emphasis',
                        strong='markdown_strong',
                        encode_entities='markdown_encode_entities',
                        codefence='markdown_codefence',
                        fences='markdown_fences',
                        list_indent='markdown_list_indent',
                        loose_tables='markdown_loose_tables',
                        spaced_tables='markdown_spaced_tables',
                        list_increment='markdown_list_increment',
                        horizontal_rule='markdown_horizontal_rule',
                        horizontal_rule_spaces='markdown_horizontal_'
                                               'rule_spaces',
                        horizontal_rule_repeat='markdown_horizontal_'
                                               'rule_repeat')
    def create_arguments(filename, file, config_file,
                         bullets: str='-',
                         closed_headings: bool=False,
                         setext_headings: bool=False,
                         emphasis: str='*',
                         strong: str='*',
                         encode_entities: bool=False,
                         codefence: str='`',
                         fences: bool=True,
                         list_indent: str='1',
                         loose_tables: bool=False,
                         spaced_tables: bool=True,
                         list_increment: bool=True,
                         horizontal_rule: str='*',
                         horizontal_rule_spaces: bool=False,
                         horizontal_rule_repeat: int=3,
                         max_line_length: int=None):
        """
        :param bullets:
            Character to use for bullets in lists. Can be "-", "*" or "+".
        :param closed_headings:
            Whether to close Atx headings or not. if true, extra # marks will
            be required after the heading. eg: `## Heading ##`.
        :param setext_headings:
            Whether to use setext headings. A setext heading uses underlines
            instead of # marks.
        :param emphasis:
            Character to wrap strong emphasis by. Can be "_" or "*".
        :param strong:
            Character to wrap slight emphasis by. Can be "_" or "*".
        :param encode_entities:
            Whether to encode symbols that are not ASCII into special HTML
            characters.
        :param codefence:
            Used to find which characters to use for code fences. Can be "`" or
            "~".
        :param fences:
            Use fences for code blocks.
        :param list_indent:
            Used to find spacing after bullet in lists. Can be "1", "tab" or
            "mixed".
            - "1" - 1 space after bullet.
            - "tab" - Use tab stops to begin a sentence after the bullet.
            - "mixed" - Use "1" when the list item is only 1 line, "tab" if it
              spans multiple.
        :param loose_tables:
            Whether to use pipes for the outermost borders in a table.
        :param spaced_tables:
            Whether to add space between pipes in a table.
        :param list_increment:
            Whether an ordered lists numbers should be incremented.
        :param markdown_horizontal_rule:
            The horizontal rule character. Can be '*', '_' or '-'.
        :param horizontal_rule_spaces:
            Whether spaces should be added between horizontal rule characters.
        :param horizontal_rule_repeat:
            The number of times the horizontal rule character will be repeated.
        :param max_line_length:
            The maximum line length allowed.
        """
        remark_configs_settings = {
            'bullet': bullets,                          # - or *
            'closeAtx': closed_headings,                # Bool
            'setext': setext_headings,                  # Bool
            'emphasis': emphasis,                       # char (_ or *)
            'strong': strong,                           # char (_ or *)
            'entities': encode_entities,                # Bool
            'fence': codefence,                         # char (~ or `)
            'fences': fences,                           # Bool
            'listItemIndent': list_indent,              # int or "tab"
                                                        # or "mixed"
            'looseTable': loose_tables,                 # Bool
            'spacedTable': spaced_tables,               # Bool
            'incrementListMarker': list_increment,      # Bool
            'rule': horizontal_rule,                    # - or * or _
            'ruleSpaces': horizontal_rule_spaces,       # Bool
            'ruleRepetition': horizontal_rule_repeat,   # int
        }
        remark_configs_plugins = {}

        if max_line_length:
            remark_configs_plugins['maximumLineLength'] = max_line_length

        config_json = json.dumps(remark_configs_settings)
        # Remove { and } as remark adds them on its own
        settings = config_json[1:-1]

        args = ['--no-color', '--quiet', '--setting', settings]

        if remark_configs_plugins:
            config_json = json.dumps(remark_configs_plugins)
            plugins = 'lint=' + config_json[1:-1]
            args += ['--use', plugins]

        return args

    def process_output(self, output, filename, file):
        stdout, stderr = output
        yield from self.process_output_corrected(stdout, filename, file,
                                                 RESULT_SEVERITY.NORMAL,
                                                 'The text does not comply'
                                                 ' to the set style.')
        yield from self.process_output_regex(stderr, filename, file,
                                             self._output_regex)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='licensecheck',
        output_format='regex',
        output_regex=r'.*: .*UNKNOWN$',
        result_message='No license found.')
class LicenseCheckBear:
    """
    Attempts to check the given file for a license, by searching the start
    of the file for text belonging to various licenses.

    For Ubuntu/Debian users, the ``licensecheck_lines`` option has to be used
    in accordance with the ``licensecheck_tail`` option.
    """
    LANGUAGES = {'All'}
    REQUIREMENTS = {
        DistributionRequirement(
            apt_get='devscripts',
            dnf='licensecheck',
            portage=None,
            zypper='devscripts',
        ),
    }
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'License'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         licensecheck_lines: int=60,
                         licensecheck_tail: int=5000):
        """
        :param licensecheck_lines:
            Specify how many lines of the file header should be parsed for
            license information. Set to 0 to parse the whole file (and ignore
            ``licensecheck_tail``).
        :param licensecheck_tail:
            Specify how many bytes to parse at end of file. Set to 0 to disable
            parsing from end of file.
        """
        return ('--lines', str(licensecheck_lines), '--tail',
                str(licensecheck_tail), filename)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='scspell',
        use_stderr=True,
        output_format='regex',
        output_regex=r'.*:(?P<line>.\d*):\s*(?P<message>.*)')
class SpellCheckBear:
    """
    Lints files to check for incorrect spellings using ``scspell``.

    scspell is a spell checker for source code.
    When applied to a code written in most popular programming languages
    while using most typical naming conventions, this algorithm will
    usually catch many errors without an annoying false positive rate.

    In an effort to catch more spelling errors, scspell is able to
    check each file against a set of dictionary words selected
    specifically for that file.

    See <https://pypi.python.org/pypi/scspell3k> for more information.
    """
    LANGUAGES = {'Natural Language'}
    REQUIREMENTS = {PipRequirement('scspell3k', '2.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/87753'
    CAN_DETECT = {'Spelling'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--report-only', filename



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.LuarocksRequirement import (
    LuarocksRequirement)


@linter(executable='luacheck',
        use_stdin=True,
        output_format='regex',
        output_regex=r'stdin:(?P<line>\d+):(?P<column>\d+)-'
                     r'(?P<end_column>\d+): '
                     r'\((?P<severity>[WE])(?P<origin>\d+)\) (?P<message>.+)')
class LuaLintBear:
    """
    Check Lua code for possible semantic problems, like unused code.

    Read more at <https://github.com/mpeterv/luacheck>.
    """

    LANGUAGES = {'Lua'}
    REQUIREMENTS = {LuarocksRequirement('luacheck')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Unreachable Code', 'Unused Code', 'Variable Misuse'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '-', '--formatter=plain', '--ranges', '--codes'



from coala_utils.string_processing import escape

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.RscriptRequirement import (
    RscriptRequirement)
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='Rscript',
        output_format='regex',
        output_regex=r'.*?:(?P<line>\d+):(?P<column>\d+): '
                     r'(?P<severity>\S+): (?P<message>.*)',
        severity_map={'style': RESULT_SEVERITY.NORMAL,
                      'warning': RESULT_SEVERITY.NORMAL,
                      'error': RESULT_SEVERITY.MAJOR},
        prerequisite_check_command=('Rscript', '-e', 'library(lintr)'),
        prerequisite_check_fail_message='R library "lintr" is not installed.')
class RLintBear:
    """
    Checks the code with ``lintr``.
    """
    LANGUAGES = {'R'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    REQUIREMENTS = {RscriptRequirement('lintr'),
                    DistributionRequirement(apt_get='r-base',
                                            version='>=3.1.1')}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ('-e', 'library(lintr)', '-e',
                'lintr::lint("' + escape(filename, '\\"') + '")')



import json

from coalib.bearlib.abstractions.Linter import linter
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result
from coalib.settings.Setting import path
from dependency_management.requirements.DistributionRequirement import (
        DistributionRequirement)


@linter(executable='tailor',
        output_format=None,
        prerequisite_check_command=('tailor', '-v'),
        prerequisite_check_fail_message='Tailor is not installed. Refer '
        'https://tailor.sh/ for installation details.')
class TailorBear:
    """
    Analyze Swift code and check for code style related
    warning messages.

    For more information on the analysis visit <https://tailor.sh/>
    """
    LANGUAGES = {'Swift'}
    REQUIREMENT = {DistributionRequirement(brew='tailor')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/45666'
    CAN_DETECT = {'Formatting'}

    severity_map = {'warning': RESULT_SEVERITY.NORMAL,
                    'error': RESULT_SEVERITY.MAJOR}

    def process_output(self, output, filename, file):
        output = json.loads(output)

        for item in output['files'][0]['violations']:
            column = (item['location']['column']
                      if 'column' in item['location'].keys() else None)

            yield Result.from_values(
                origin='{} ({})'.format(self.name, item['rule']),
                message=item['message'],
                file=filename,
                line=item['location']['line'],
                column=column,
                severity=self.severity_map[item['severity']])

    @staticmethod
    def create_arguments(filename, file, config_file,
                         max_line_length: int=79,
                         max_class_length: int=0,
                         max_closure_length: int=0,
                         max_file_length: int=0,
                         max_function_length: int=0,
                         max_name_length: int=0,
                         max_struct_length: int=0,
                         min_name_length: int=1,
                         tailor_config: path=''):
        """
        Bear configuration arguments.
        Using '0' will disable the check.

        :param max_line_length:     maximum number of characters in a Line
                                    <0-999>.
        :param max_class_length:    maximum number of lines in a Class <0-999>.
        :param max_closure_length:  maximum number of lines in a Closure
                                    <0-999>.
        :param max_file_length:     maximum number of lines in a File <0-999>.
        :param max_function_length: maximum number of lines in a Function
                                    <0-999>.
        :param max_name_length:     maximum length of Identifier name <0-999>.
        :param max_struct_length:   maximum number od lines in a Struct
                                    <0-999>.
        :param min_name_length:     minimum number of characters in Identifier
                                    name <1-999>.
        :param tailor_config:       path to Tailor configuration file.
        """
        args = ('--format=json',
                '--max-line-length', str(max_line_length),
                '--max-class-length', str(max_class_length),
                '--max-closure-length', str(max_closure_length),
                '--max-file-length', str(max_file_length),
                '--max-function-length', str(max_function_length),
                '--max-name-length', str(max_name_length),
                '--max-struct-length', str(max_struct_length),
                '--min-name-length', str(min_name_length))
        if tailor_config:
            args += ('--config=' + tailor_config,)
        return args + (filename,)
